/*
 * Copyright 2022 KylinSoft Co., Ltd.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with thi1s program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include "applauncher.h"
#include <QtConcurrent>
#include <QProcess>
#include <QStandardPaths>
#include <QFileSystemWatcher>

extern "C" {
#include <unistd.h>
#include <fcntl.h>
#include <gio/gdesktopappinfo.h>
#include <gio/gappinfo.h>
}

AppLauncher::AppsInfo AppLauncher::m_appsInfo;
AppLauncher::AppLauncher(QObject *parent)
    : QObject(parent)
    , m_isTabletMode(false), m_isWakeupHoverButton(false)
#ifdef INVISIBLE_ENABLE
    , m_timer(new QTimer(this))
    , currentCounter(0)
#endif
    , m_powerSavingEnabled(false)
    , m_cgroupInterface(new com::kylin::CGroup::AppManager("com.kylin.CGroup", "/com/kylin/CGroup/AppManager",
                                                            QDBusConnection::systemBus(), this))
    , m_kwinInterface(new org::ukui::KWin("org.ukui.KWin", "/KWin",
                                          QDBusConnection::sessionBus(), this))
    , m_statusManagerInterface(new com::kylin::statusmanager::interface("com.kylin.statusmanager.interface", "/",
                                                                        QDBusConnection::sessionBus(), this))
    , m_appLauncherDaemon(new AppLauncherDaemon(this))
    , m_whiteListManager(new WhiteListManager(this))
    , m_desktopFileWatcher(new QFileSystemWatcher(this))
    , m_polkitWindowPid(0)
{
    if (!m_statusManagerInterface->isValid()) {
        m_isTabletMode = false;
        qWarning() << "com.kylin.statusmanager.interface is not valid.";
    } else {
        m_isTabletMode = m_statusManagerInterface->get_current_tabletmode();
    }

#ifdef QT_DEBUG
//    m_isTabletMode = true;
#endif
    qDebug() << "current tablet mode is: " << m_isTabletMode;
    initDesktopFileWatcher();
    initConnections();

    QTimer::singleShot(1, [this] {
        auto reply = m_cgroupInterface->DeleteCGroup();
        reply.waitForFinished();
        if (reply.isError()) {
            qWarning() << "Delete CGroup failed: " << reply.error();
        }
        reply = m_cgroupInterface->CreateCGroup();
        if (reply.isError()) {
            qWarning() << "Create CGroup failed: " << reply.error();
        }
    });
}

bool AppLauncher::LaunchApp(const QString &desktopFile)
{
    return doLaunchApp(desktopFile);
}

bool AppLauncher::ActiveProcessByWid(const WindowId &wid)
{
    return doActiveProcessByWid(wid);
}

bool AppLauncher::LaunchAppWithArguments(const QString &desktopFile, const QStringList &args)
{
    return doLaunchApp(desktopFile, args);
}

bool AppLauncher::Open(const QString &fileName)
{
    if (fileName.isEmpty()) {
        qWarning() << __FUNCTION__ << "fileName is NULL!";
        return false;
    }
    QUrl url(fileName);
    QFile file(fileName);
    if (!url.scheme().isEmpty()) {
        // 传入的参数为url类型
        return LaunchDefaultAppWithUrl(fileName);
    }
    if (!file.exists()) {
        qWarning() << __FUNCTION__ << fileName << "is invalid!";
        return false;
    }
    QVector<QStringList> launchApplists = getLaunchAppLists(fileName);
    if (launchApplists.isEmpty()) {
        qWarning() << __FUNCTION__ << fileName << "is not have apps to launch!";
        return false;
    }
    if (!launchApplists.first().isEmpty()) {
        // 选用默认应用打开文件
        for (int i=0; i<launchApplists.first().count(); ++i) {
            if (LaunchAppWithArguments(launchApplists.first().at(i), QStringList() << fileName)) {
                return true;
            }
        }
    }
    // 选用推荐应用列表打开文件，直到应用打开文件为止
    for (int i=0; i<launchApplists.last().count(); ++i) {
        if (LaunchAppWithArguments(launchApplists.last().at(i), QStringList() << fileName)) {
            return true;
        }
    }

    return false;
}

bool AppLauncher::LaunchDefaultAppWithUrl(const QString &url)
{
    int ret = g_app_info_launch_default_for_uri(url.toUtf8().constData(),
                                                nullptr,
                                                nullptr);
    qDebug() << __func__ << url << ret;
    if (ret) {
        return true;
    }

    QUrl url_(url);
    if (url_.isEmpty() || !url_.isValid()) {
        return false;
    }
    QProcess p;
    if (url_.isLocalFile()) {
        QString fileName = url_.toLocalFile();
        qDebug ()<<"---fileName---" << fileName;
        p.start("xdg-mime", QStringList() << "query" << "filetype" << fileName);
        p.waitForFinished(3000);
        if (p.error() != QProcess::UnknownError) {
            qWarning() << __FUNCTION__ << "query filetype for "
                       << fileName << "failed!" << p.errorString();
            return false;
        }
        QString fileType = p.readAll();
        if (fileType.endsWith("\n")) {
            fileType.remove("\n");
        }
        qDebug ()<<"---fileType---" << fileType;
        p.start("xdg-mime", QStringList() << "query" << "default" << fileType);
        p.waitForFinished(3000);
        if (p.error() != QProcess::UnknownError) {
            qWarning() << __FUNCTION__ << "query default for "
                       << fileType << "failed!" << p.errorString();
            return false;
        }

        QString desktopFile = p.readAll();
        if (desktopFile.endsWith("\n")) {
            desktopFile.remove("\n");
        }
        return doLaunchApp(desktopFile, QStringList() << fileName);
    } else {
        QString scheme = url_.scheme();
        QString mimeType = "x-scheme-handler/" + scheme;
        qDebug ()<<"---mimeType---" << mimeType;
        p.start("xdg-mime", QStringList() << "query" << "default" << mimeType);
        p.waitForFinished(3000);
        if (p.error() != QProcess::UnknownError) {
            qWarning() << __FUNCTION__ << "query default for "
                       << mimeType << "failed!" << p.errorString();
            return false;
        }
        QString desktopFile = p.readAll();
        if (desktopFile.endsWith("\n")) {
            desktopFile.remove("\n");
        }
        return doLaunchApp(desktopFile, QStringList() << url);
    }
    return true;
}

QString AppLauncher::AppDesktopFileNameByPid(qint64 pid)
{
    auto it = m_appsInfo.constBegin();
    while (it != m_appsInfo.constEnd()) {
        auto processIt = it->constBegin();
        while (processIt != it->constEnd()) {
            if (processIt->pid() == pid) {
                qDebug() << __FUNCTION__ << processIt->desktopFileName() << pid;
                return processIt->desktopFileName();
            }
            ++ processIt;
        }
        ++ it;
    }
    return QString();
}

QString AppLauncher::AppDesktopFileNameByWid(qint64 wid)
{
    qDebug() << __FUNCTION__ << "--" << wid;
    auto it = m_appsInfo.constBegin();
    while (it != m_appsInfo.constEnd()) {
        auto processIt = it->constBegin();
        while (processIt != it->constEnd()) {
            for (auto &appWid : processIt->wids()) {
                if (appWid == wid) {
                    return processIt->desktopFileName();
                }
            }
            ++ processIt;
        }
        ++ it;
    }
    return QString();
}

QVector<QStringList> AppLauncher::RecommendAppLists(const QString &fileName)
{
    if (fileName.isEmpty()) {
        qWarning() << __FUNCTION__ << fileName << "is NULL!";
        return QVector<QStringList>();
    }
    QUrl url(fileName);
    QFile file(fileName);
    if (url.scheme().isEmpty()) {
        if (file.exists() && url.isRelative()) {
            // fileName为本地文件类型
            return getLaunchAppLists(fileName);
        }
    } else {
        QString localFileName = url.toLocalFile();
        if (QFile(localFileName).exists()) {
            // fileName为url类型
            return getLaunchAppLists(localFileName);
        }
    }
    qWarning() << fileName << "is not exists!";
    return QVector<QStringList>();
}

QVector<QStringList> AppLauncher::getLaunchAppLists(const QString &fileName)
{
    QVector <QStringList> appLists;
    QStringList defaultLists;
    QStringList recommendLists;
    // 获取文件的mime-type
    QMimeDatabase fileDate;
    QMimeType fileMimeType = fileDate.mimeTypeForFile(fileName);
    // 获取默认打开应用
    GAppInfo *appInfo = g_app_info_get_default_for_type(fileMimeType.name().toUtf8().constData(), false);

    if (appInfo == nullptr) {
        qWarning() << "appInfo is NULL, localFilePath not have default launch app!";
        appLists.push_back(QStringList());
    } else {
        const char *defaultAppName = g_app_info_get_id(appInfo);
        defaultLists.append(QString::fromUtf8(defaultAppName));
        appLists.push_back(defaultLists);
    }
    // 获取推荐打开应用列表
    GList *app_infos = g_app_info_get_recommended_for_type(fileMimeType.name().toUtf8().constData());
    GList *l = app_infos;
    if (app_infos == nullptr) {
        qWarning() << "app_infos is NULL, localFilePath not have recommend launch apps!";
        appLists.push_back(QStringList());
    } else {
        while (l != nullptr) {
            auto app_info = static_cast <GAppInfo*>(l->data);
            if (app_info == nullptr) {
                qWarning() << "app_info is nullptr!";
                recommendLists.append("");
                l = l->next;
                continue;
            } else {
                const char *appDesktopName = g_app_info_get_id(app_info);
                if (defaultLists.contains(QString::fromUtf8(appDesktopName))) {
                    l = l->next;
                    continue;
                }
                recommendLists.append(QString::fromUtf8(appDesktopName));
                l = l->next;
            }
            g_object_unref(app_info);
        }
        appLists.push_back(recommendLists);
    }
    g_list_free(l);
    g_list_free(app_infos);
    return appLists;
}

bool AppLauncher::isWakeupHoverButtonApp(const QString &fullDesktopName)
{
    if (fullDesktopName.isEmpty()) {
        qWarning() << __FUNCTION__ << "fullDesktopName is null!";
        return false;
    }
    if (fullDesktopName.contains("wps-office")) {
        return true;
    }
    return false;
}

bool AppLauncher::isWakeupHoverButtonApp(const int &pid)
{
    if (pid == 0 || pid < 0) {
        qWarning() << __FUNCTION__ << "pid is not valid!";
        return false;
    }
    if (ProcessInfo::cmdline(pid).contains("wps-office")) {
        return true;
    }
    return false;
}

/// 当 \param enable 设为 true (即开启省电模式)时，
/// 将 \memberof status 为 'InvisibleProcess' 的进程移动至 'CachedProcess' cgroup 分组
/// 关闭省电模式时，将所有的 'CachedProcess' 分组的应用移动至 'InvisibleProcess' 分组
void AppLauncher::setPowerSavingModeEnable(bool enable)
{
    if (m_powerSavingEnabled == enable) {
        return;
    }

    m_powerSavingEnabled = enable;
}

void AppLauncher::killCachedApp()
{
#if 0
    for (const auto &desktopFile : m_appsInfo.keys()) {
        if (m_appsInfo[desktopFile].status == CachedProcess) {
            qint64 pid = m_appsInfo[desktopFile].pid;
            // 先移动至 InvisibleProcess 分组，才能将其 kill 掉,cached组中进程被冻结
            updateAppStatus(pid, InvisibleProcess);
            bool ret = QProcess::startDetached("kill", QStringList() << "-9" << QString::number(pid));
            if (ret) {
                m_appsInfo.remove(desktopFile);
            }
        }
    }
#endif
}

void AppLauncher::tabletModeChanged(bool tabletMode)
{
    qDebug() << "tablet changed: " << tabletMode;
    if (m_isTabletMode == tabletMode) {
        return;
    }
    m_isTabletMode = tabletMode;

#ifdef INVISIBLE_ENABLE
    if (tabletMode) {
        m_timer->start(1000);
    } else {
        m_timer->stop();
        m_timerContainer.clear();
        currentCounter = 0;
    }
#endif
    updateAppCGroupStatus();
}

void AppLauncher::activeWindowChanged(const WindowId &window)
{
    if (!m_isTabletMode) {
        return;
    }
    static quint32 preActivePid = 0;
    //quint32 wid = m_kwinInterface->getActiveWindowId();
    //quint32 pid = m_kwinInterface->getWindowPid(wid);
    quint32 pid = WindowManager::getPid(window);

    updateAppCGroupStatus(pid, ProcessInfo::ForegroundProcess);
    if (preActivePid != 0 && preActivePid != pid) {
        updateAppCGroupStatus(preActivePid, ProcessInfo::CachedProcess);
    }
    preActivePid = pid;

    // 活动窗口为Wps等类应用时唤醒悬浮按钮
    if (!isWakeupHoverButtonApp(pid) && m_isWakeupHoverButton == true) {
        m_isWakeupHoverButton = false;
        launch("ukui-vkeyboard-hoverbutton", QStringList() << "-q");
    }
    if (isWakeupHoverButtonApp(pid) && m_isWakeupHoverButton == false) {
        m_isWakeupHoverButton = true;
        launch("ukui-vkeyboard-hoverbutton");
    }
}

#ifdef INVISIBLE_ENABLE
void AppLauncher::timeOut()
{
    if (m_timerContainer.isEmpty()) {
        return;
    }
    ++ currentCounter;

    handleTimeOut();

    if (currentCounter == UINT64_MAX) {
        currentCounter = 0;
    }
}

void AppLauncher::handleTimeOut()
{
    if (m_timerContainer.isEmpty()) {
        return;
    }
    if (m_timerContainer.first().deadCounter == currentCounter) {
        QString desktopFile = m_timerContainer.first().desktopFile;
        m_timerContainer.removeFirst();
        qDebug() << "move process to CachedProcess group: " << desktopFile;
//        updateAppStatus(desktopFile, CachedProcess);
        handleTimeOut();
    } else {
        return;
    }
}
#endif

void AppLauncher::aboutToQuitApp()
{
    qDebug() << "about to quit app";
    moveCachePidToGroup(ProcessInfo::ForegroundProcess);
    m_cgroupInterface->DeleteCGroup();
}

void AppLauncher::windowAdded(const WindowId &window)
{
    windowChanged(window, true);
}

void AppLauncher::windowRemoved(const WindowId &window)
{
    windowChanged(window, false);
}

void AppLauncher::initDesktopFileWatcher()
{
    QStringList desktopDirs = QStandardPaths::standardLocations(QStandardPaths::ApplicationsLocation);
    qDebug() << desktopDirs;
    auto appExecFunction = [desktopDirs, this]() {
        for (const auto &desktopDir : desktopDirs) {
            updateDesktopExecsInfo(desktopDir);
        }
    };
    QtConcurrent::run(appExecFunction);
    m_desktopFileWatcher->addPaths(desktopDirs);
}

void AppLauncher::initConnections()
{
    connect(m_statusManagerInterface, &com::kylin::statusmanager::interface::mode_change_signal,
            this, &AppLauncher::tabletModeChanged);

#ifdef INVISIBLE_ENABLE
    connect(m_timer, &QTimer::timeout, this, &AppLauncher::timeOut);
#endif

    connect(WindowManager::self(), &WindowManager::activeWindowChanged,
            this, &AppLauncher::activeWindowChanged);

    connect(m_appLauncherDaemon, &AppLauncherDaemon::aboutToQuitAppManager,
            this, &AppLauncher::aboutToQuitApp);

    connect(WindowManager::self(), &WindowManager::windowAdded,
            this, &AppLauncher::windowAdded);

    connect(WindowManager::self(), &WindowManager::windowRemoved,
            this, &AppLauncher::windowRemoved);

    connect(m_desktopFileWatcher, &QFileSystemWatcher::directoryChanged,
            this, &AppLauncher::updateDesktopExecsInfo);
}

bool AppLauncher::updateAppCGroupStatus(int pid, ProcessInfo::AppCGroup status)
{
    auto it = m_appsInfo.begin();
    while (it != m_appsInfo.end()) {
        auto appListIt = std::find_if(it->begin(), it->end(), [pid](const ProcessInfo &appInfo) {
           return appInfo.pid() == pid;
       });
       if (appListIt != it->end()) {
           bool isExist = m_whiteListManager->isExists(it.key(), "AppManager");
           qDebug() << "m_whiteListManager" << it.key() << isExist << appListIt->lockingCgroup() << appListIt->appType();
           if (appListIt->lockingCgroup() || appListIt->appType() == ProcessInfo::Kmre) {
               return false;
           }
           if (isExist) {
               if (appListIt->cgroup() == ProcessInfo::ForegroundProcess) {
                   return false;
               } else {
                   appListIt->setCGroup(ProcessInfo::ForegroundProcess);
                   QList<qint64> childrenPids = appListIt->childrenPids();
                   childrenPids << appListIt->getTasks();
                   qDebug() << "move children pid" << childrenPids;
                   m_cgroupInterface->MoveProcess(pid, ProcessInfo::ForegroundProcess);
                   for (const auto &childPid : childrenPids) {
                       auto reply = m_cgroupInterface->MoveProcess(childPid, ProcessInfo::ForegroundProcess);
                       reply.waitForFinished();
                   }
                   return true;
               }
           }

           appListIt->setCGroup(status);
           QList<qint64> childrenPids = appListIt->childrenPids();
           childrenPids << appListIt->getTasks();
           qDebug() << "move children pid" << childrenPids;
           m_cgroupInterface->MoveProcess(pid, status);
           for (const auto &childPid : childrenPids) {
               auto reply = m_cgroupInterface->MoveProcess(childPid, status);
               reply.waitForFinished();
           }
           return true;
       }
        ++ it;
    }
    return false;
}

void AppLauncher::updateAppCGroupStatus()
{
    if (!m_isTabletMode) {
        for (auto &appList : m_appsInfo) {
            for (auto &app : appList) {
                app.setCGroup(ProcessInfo::ForegroundProcess);
            }
        }
        moveCachePidToGroup(ProcessInfo::ForegroundProcess);
        qDebug() << __FUNCTION__ << m_isTabletMode << m_appsInfo;
    } else {
        //uint wid = m_kwinInterface->getActiveWindowId();
        //uint pid = m_kwinInterface->getWindowPid(wid);
        WindowId window = WindowManager::currentActiveWindow();
        uint pid = WindowManager::getPid(window);

        auto appInfoIt = m_appsInfo.begin();
        while (appInfoIt != m_appsInfo.end()) {
            bool isExist = m_whiteListManager->isExists(appInfoIt.key(), "AppManager");
            if (isExist || isKmreApp(appInfoIt.key())) {
                ++ appInfoIt;
                continue;
            }
            auto appIt = appInfoIt->begin();
            while (appIt != appInfoIt->end()) {
                if (appIt->lockingCgroup() || appIt->appType() == ProcessInfo::Kmre) {
                    ++ appIt;
                    continue;
                }
                auto allPids = appIt->childrenPids();
                allPids.push_front(appIt->pid());
                qDebug() << "all pids " << allPids << pid;
                if (appIt->pid() == pid) {
                    appIt->setCGroup(ProcessInfo::ForegroundProcess);
                    for (const auto &pid : allPids) {
                        m_cgroupInterface->MoveProcess(pid, ProcessInfo::ForegroundProcess);
                    }
                } else {
                    appIt->setCGroup(ProcessInfo::CachedProcess);
                    for (const auto &pid : allPids) {
                        m_cgroupInterface->MoveProcess(pid, ProcessInfo::CachedProcess);
                    }
                }
                ++ appIt;
            }
            ++ appInfoIt;
        }
        qDebug() << __FUNCTION__ << m_isTabletMode << m_appsInfo;
    }
}

QPair<QString, QStringList> AppLauncher::desktopFileExec(const QString &desktopFileName)
{
    if (desktopFileName.isEmpty()) {
        qWarning() << __FUNCTION__ << "desktopFileName is NULL！！！";
        return QPair<QString, QString>();
    } else {
        QString fullDesktopName = desktopFullFilename(desktopFileName);
        KDesktopFile desktop(fullDesktopName);
        QString exec = desktop.entryMap("Desktop Entry").value("Exec");

        exec.remove(QRegExp("%."));
        exec.remove(QRegExp("^\""));
        exec.remove(QRegExp(" *$"));
        exec.remove(QRegExp("\""));

        QStringList execArgs = exec.split(" ");
        if (execArgs.count() > 0) {
            exec = execArgs.first();
        }
        return { exec, execArgs };
    }
}

bool AppLauncher::isKmreApp(const QString &desktopFile)
{
    if (desktopFile.isEmpty()) {
        qWarning() << __FUNCTION__ << "desktopFileName is NULL！！！";
        return false;
    } else {
        QString fullDesktopName = desktopFullFilename(desktopFile);
        QScopedPointer<QSettings> desktop(new QSettings(fullDesktopName, QSettings::IniFormat));
        desktop->setIniCodec(QTextCodec::codecForName("UTF-8"));
        QString categories = desktop->value("Desktop Entry/Categories").toString();
        categories.remove(QRegExp("%."));
        categories.remove(QRegExp("^\""));
        categories.remove(QRegExp(" *$"));
        return categories.contains("Android") || categories.contains("Apk");
    }
    return false;
}

bool AppLauncher::hasWindow(const QString &desktopFile)
{
    QString fullDesktopName = desktopFullFilename(desktopFile);
    QScopedPointer<QSettings> desktop(new QSettings(fullDesktopName, QSettings::IniFormat));
    desktop->setIniCodec(QTextCodec::codecForName("UTF-8"));
    QString keywords = desktop->value("Desktop Entry/Keywords").toString();
    keywords.remove(QRegExp("%."));
    keywords.remove(QRegExp("^\""));
    keywords.remove(QRegExp(" *$"));
    if (keywords.contains("screenshot") || keywords.contains("capture") || keywords.contains("shutter")) {
        return false;
    }

    return true;
}

#ifdef INVISIBLE_ENABLE
/// 关于定时器的过程：
/// 平板模式下，将会开启一个1s的定时器 \memberof m_timer，每次超时都会将 \memberof currentCounter +1，
/// 当进程处于‘不可见进程’状态时，为进程添加一个 \struct Timer，包含了超时时的 \memberof currentCounter
/// 值，并将该 Timer 添加到 \memberof m_timerContainer 中去，每次 m_timer 超时，都会去 m_timerContainer
/// 中找到当前与 currentCounter 相等的数值，并对相应的进程进程处理。
void AppLauncher::addTimer(const QString &desktopFile)
{
    qDebug() << "addTimer: " << desktopFile;
    deleteTimer(desktopFile);
    m_timerContainer.append({ currentCounter + interval, desktopFile });
}

///
void AppLauncher::deleteTimer(const QString &desktopFile)
{
    for (auto timer : m_timerContainer) {
        if (timer.desktopFile == desktopFile) {
            m_timerContainer.removeOne(timer);
            qDebug() << "deleteTimer: " << desktopFile;
            return;
        }
    }
}
#endif

QString AppLauncher::desktopFullFilename(const QString &desktopFileName)
{
    QFile desktopFile(desktopFileName);
    if (desktopFile.exists()) {
        return desktopFileName;
    }
    QStringList desktopfilePaths = QStandardPaths::standardLocations(QStandardPaths::ApplicationsLocation);
    for (auto path : desktopfilePaths) {
        if (desktopFileName.contains(path)) {
            return desktopFileName;
        } else if (QFile::exists(path + "/" + desktopFileName)) {
            return path + "/" + desktopFileName;
        }
    }
    return QString();
}

bool AppLauncher::doLaunchApp(const QString &desktopFile, const QStringList &args)
{
    QString fullDesktopName = desktopFullFilename(desktopFile);
    if (fullDesktopName.isEmpty()) {
        qWarning() << "Launch app failed, the dekstop file is " << desktopFile;
        return false;
    }
    if (!KDesktopFile::isDesktopFile(fullDesktopName)) {
        return false;
    }
    qDebug() << "fullDesktopName" << fullDesktopName;

#if 1 //crashed: corrupted double-linked list. mybe a bug
    QScopedPointer<KDesktopFile> file(new KDesktopFile(fullDesktopName));
    if (!file->hasApplicationType() && !file->tryExec()) {
        return false;
    }
#endif

    // not full path file name
    QString dekstopFilename = fullDesktopName.split("/").last();

    bool launchResult = false;
    d.fork = 0;
    QStringList startUpArgs = args;
    QPair<QString, QStringList> execArgs;
    if (m_desktopExec.contains(dekstopFilename)) {
        execArgs = m_desktopExec.value(dekstopFilename);
    } else {
        execArgs = desktopFileExec(desktopFile);
    }
    if (args.isEmpty()) {
        startUpArgs = execArgs.second;
    }
    // 文件打开方式选择kmre应用打开时，只能启动应用，不能打开文件
    if (isKmreApp(desktopFile)) {
        startUpArgs = execArgs.second;
        for (int i=0; i<args.count(); ++i) {
            startUpArgs.push_back(args.at(i));
        }
    }
    // 打开的应用exec包含sh -c的需要将整个exec字段作为一个字符串传入，而不是单个参数传入
    if (execArgs.second.isEmpty()) {
        qWarning() << "ExecArgs.second is Empty";
    }
    if (execArgs.second.count() > 1) {
        if (execArgs.second[0] == ("sh") && execArgs.second[1] == ("-c")) {
            QStringList execList;
            QStringList execStr;
            execList << "sh" << "-c";
            for (int i=2; i<execArgs.second.count(); ++i) {
                execStr << execArgs.second.at(i);
            }
            execStr << args.join(" ");
            execList << execStr.join(" ");
            startUpArgs = execList;
        }
    }
    if (!startUpArgs.startsWith(execArgs.first)) {
        startUpArgs.prepend(execArgs.first);
    }
    QString exec = execArgs.first;
    ProcessInfo appInfo;
    appInfo.setAppStatus(ProcessInfo::Launching);
    appInfo.setArgs(startUpArgs);
    appInfo.setDesktopFileName(dekstopFilename);

    if (!hasWindow(desktopFile) || kFreeApps.contains(dekstopFilename)) {
        return launchDesktopApp(fullDesktopName, args, exec, startUpArgs);
    }

    qDebug() << exec << startUpArgs;
    // kmre应用以app的名称作为唯一标识，比如com.sina.weibo
    if (isKmreApp(desktopFile)) {
        launchResult = launchDesktopApp(fullDesktopName, args, exec, startUpArgs);
        if (startUpArgs.size() < 2) {
            qWarning() << "Kmre app desktop args error, " << startUpArgs;
            return false;
        }
        appInfo.setAppType(ProcessInfo::Kmre);
        QList<ProcessInfo> appList;
        appList.push_back(appInfo);
        m_appsInfo[startUpArgs.at(1)] = appList;
        return launchResult;
    }

    appInfo.setAppType(ProcessInfo::Normal);
    if (!m_isTabletMode) {
        m_appsInfo[dekstopFilename].push_back(appInfo);
        launchResult = launchDesktopApp(fullDesktopName, args, exec, startUpArgs);
    } else {
        auto appRuningStatus = appStatus(dekstopFilename);
        switch (appRuningStatus) {
        case ProcessInfo::None: {
            m_appsInfo[dekstopFilename].push_back(appInfo);
            launchResult = launchDesktopApp(fullDesktopName, args, exec, startUpArgs);
            break;
        }
        case ProcessInfo::Launching: {
            if (kNeedingPolkitWindowApps.contains(dekstopFilename)) {
                launchResult = activePolkitWindow();
                if (!launchResult) {
                    launchResult = launchDesktopApp(fullDesktopName, args, exec, startUpArgs);
                }
            }
            if (!tryToActiveExistWindow(startUpArgs) &&
                    launchingTimeout(dekstopFilename, startUpArgs)) {
                launchResult = launchDesktopApp(fullDesktopName, args, exec, startUpArgs);
            }
            break;
        }
        case ProcessInfo::Running: {
            if (!activeAppWindow(dekstopFilename)) {
                qWarning() << "active app window failed," << dekstopFilename << startUpArgs;
                m_appsInfo[dekstopFilename].push_back(appInfo);
                launchResult = launchDesktopApp(fullDesktopName, args, exec, startUpArgs);
            }
            break;
        }
        default:
            break;
        }
    }
    if (launchResult && d.fork > 0) {
        m_appsInfo[dekstopFilename].last().setLaunchPid(d.fork);
    }
    // Wps等类应用需要唤醒悬浮按钮
    if (isWakeupHoverButtonApp(fullDesktopName) && m_isWakeupHoverButton == false) {
        m_isWakeupHoverButton = true;
        launch("ukui-vkeyboard-hoverbutton", QStringList());
    }

    return launchResult;
}

bool AppLauncher::activeAppWindow(const QString &desktopFile, const QStringList &args)
{
    if (!m_appsInfo.contains(desktopFile)) {
        return false;
    }
    auto appList = m_appsInfo.value(desktopFile);
    auto it = appList.constEnd() - 1;
    while (it != appList.constBegin() - 1) {
        if (it->appStatus() == ProcessInfo::Running &&
                it->getArgs() == args &&
                !it->wids().isEmpty()) {
    	    // kysdk的activateWindow接口没有返回值
            WindowManager::activateWindow(it->wids().last());
            bool actived = true;
            if (actived) {
                return true;
            }
        }
        -- it;
    }
    return false;
}

bool AppLauncher::activeAppWindow(const QString &desktopFile)
{
    if (!m_appsInfo.contains(desktopFile)) {
        return false;
    }

    auto appList = m_appsInfo.value(desktopFile);
    if (appList.isEmpty()) {
        qWarning() << desktopFile << "info is empty";
        return false;
    }

    auto appListIt = appList.constEnd() - 1;
    while (appListIt != appList.constBegin() - 1) {
        if (appListIt->appStatus() == ProcessInfo::Running) {
            bool ret = m_kwinInterface->activeWindowByPid(appListIt->pid());
            // kysdk的activateWindow接口没有返回值
            WindowManager::activateWindow(appListIt->wids().last());
            ret = true;
            if (!ret) {
                qWarning() << "Active window by pid " << appListIt->pid() << "failed.";
                -- appListIt;
                continue;
            } else {
                return true;
            }
        }
        -- appListIt;
    }

    return false;
}

bool AppLauncher::activePolkitWindow()
{
    WindowManager::activateWindow(m_polkitWindowId);
    return true;
}

ProcessInfo::AppStatus AppLauncher::appStatus(const QString &desktopFile, const QStringList &args)
{
    if (!m_appsInfo.contains(desktopFile)) {
        return ProcessInfo::None;
    }
    auto appList = m_appsInfo.value(desktopFile);
    auto it = std::find_if(appList.constBegin(), appList.constEnd(), [args](const ProcessInfo &appInfo) {
        return appInfo.getArgs() == args;
    });
    if (it != appList.constEnd()) {
        return it->appStatus();
    }
    return ProcessInfo::None;
}

ProcessInfo::AppStatus AppLauncher::appStatus(const QString &desktopFile)
{
    if (!m_appsInfo.contains(desktopFile)) {
        return ProcessInfo::None;
    }

    auto appList = m_appsInfo.value(desktopFile);
    if (appList.isEmpty()) {
        qWarning() << desktopFile << "info is empty";
        return ProcessInfo::None;
    }

    // 先找Running状态的，再找Launching状态的
    bool hasLaunchingStatus = false;
    auto appListIt = appList.constEnd() - 1;
    while (appListIt != appList.constBegin() - 1) {
        if (appListIt->appStatus() == ProcessInfo::Running) {
            return ProcessInfo::Running;
        }
        if (appListIt->appStatus() == ProcessInfo::Launching) {
            hasLaunchingStatus = true;
        }
        -- appListIt;
    }

    if (hasLaunchingStatus) {
        return ProcessInfo::Launching;
    }

    return ProcessInfo::None;
}

bool AppLauncher::tryToActiveExistWindow(const QStringList &args)
{
    if (args.isEmpty()) {
        return false;
    }
    for (auto &appList : m_appsInfo) {
        for (auto &app : appList) {
            if (app.getArgs() == args) {
                WindowManager::activateWindow(app.wids().last());
                bool ret = true;
                if (ret) {
                    return true;
                } else {
                    qWarning() << "KWin active window by pid failed, " << app.pid();
                    return false;
                }
            }
        }
    }
    return false;
}

bool AppLauncher::launchingTimeout(const QString &desktop, const QStringList &args)
{
    if (!m_appsInfo.contains(desktop)) {
        return false;
    }
    auto &appList = m_appsInfo[desktop];
    for (auto &app : appList) {
        if (app.getArgs() == args) {
            if (QDateTime::currentSecsSinceEpoch() - app.launchTimestamp() > 10) {
                appList.removeOne(app);
                return true;
            } else {
                return false;
            }
        }
    }
    return true;
}

void AppLauncher::updateDesktopExecsInfo(const QString &path)
{
    QDir dir(path);
    if (!dir.exists()) {
        return;
    }

    QStringList desktopFiles = dir.entryList(QStringList() << "*.desktop");
    for (const auto &desktopFileName : desktopFiles) {
        auto exec = desktopFileExec(path + "/" + desktopFileName);
        {
            QMutexLocker locker(&m_mutex);
            if (!m_desktopExec.contains(desktopFileName)) {
                m_desktopExec[desktopFileName] = exec;
            }
        }
    }
}

void AppLauncher::updateWhitelistCGroupStatus(const QString &desktopName)
{
    if (!m_appsInfo.contains(desktopName)) {
        qWarning() << __FUNCTION__ << "m_appsInfo is not contains " << desktopName;
        return;
    }
    auto appList = m_appsInfo.value(desktopName);
    for (auto &appInfo : appList) {
        if (appInfo.cgroup() != ProcessInfo::ForegroundProcess) {
            appInfo.setCGroup(ProcessInfo::ForegroundProcess);
            QList<qint64> childrenPids = appInfo.childrenPids();
            childrenPids << appInfo.getTasks();
            qDebug() << __FUNCTION__ << "move children pid" << childrenPids;
            m_cgroupInterface->MoveProcess(appInfo.pid(), ProcessInfo::ForegroundProcess);
            for (const auto &childPid : childrenPids) {
                auto reply = m_cgroupInterface->MoveProcess(childPid, ProcessInfo::ForegroundProcess);
                reply.waitForFinished();
            }
            return;
        }
    }
}

bool AppLauncher::launch(const QString &execPath, const QStringList &args)
{
    if (pipe(d.fd) < 0) {
        qWarning() << "pipe() failed," << strerror(errno);
        d.fork = 0;
        return false;
    }

    {
        if (d.argv == nullptr) {
            d.argv = new char*[args.count() + 1];
            for (int i=0; i<args.count(); ++i) {
                QByteArray *arg = new QByteArray;
                *arg = args.at(i).toLocal8Bit();
                d.argv[i] = arg->data();
                qDebug() << "arg" << arg << d.argv[i] << arg->data();
            }
            d.argv[args.count()] = nullptr;
        }
    }

    d.fork = fork();
    switch (d.fork) {
    case -1: {
        qWarning() << execPath << "fork failed, " << strerror(errno);
        close(d.fd[0]);
        close(d.fd[1]);
        d.fork = 0;
        return false;
    }
    // child process
    case 0: {
        close(d.fd[0]);
        d.result = 2; // Try execing
        write(d.fd[1], &d.result, 1);

        // We set the close on exec flag.
        // Closing of d.fd[1] indicates that the execvp succeeded!
        fcntl(d.fd[1], F_SETFD, FD_CLOEXEC);

        QByteArray executable;
        executable.append(execPath);
        qWarning() << "exec" << executable.toStdString().c_str();
        if (!executable.isEmpty()) {
            qDebug() << execvp(executable.constData(), d.argv);
            delete [] d.argv;
            d.argv = nullptr;
        }
        qWarning() << "Launch error" << strerror(errno);
        d.result = 1; // Error
        write(d.fd[1], &d.result, 1);
        close(d.fd[1]);
        exit(255);
        break;
    }

    default: {
        qDebug() << "child process pid " << d.fork;
        close(d.fd[1]);
        bool exec = false;
        for (;;) {
            qDebug() << "start read....";
            d.n = read(d.fd[0], &d.result, 1);
            if (d.n == 1) {
                if (d.result == 2) {
                    exec = true;
                    continue;
                }
                if (d.result == 3) {
                    int l = 0;
                    d.n = read(d.fd[0], &l, sizeof(int));
                    if (d.n == sizeof(int)) {
                        QByteArray tmp;
                        tmp.resize(l + 1);
                        d.n = read(d.fd[0], tmp.data(), l);
                        tmp[l] = 0;
                        if (d.n == l) {
                            d.errorMsg = tmp;
                        }
                    }
                }
                qWarning() << "process finished" << errno << "str error " << strerror(errno);
                // Finished
                break;
            }
            if (d.n == -1) {
                if (errno == ECHILD) {  // a child died.
                    continue;
                }
                if (errno == EINTR || errno == EAGAIN) { // interrupted or more to read
                    continue;
                }
            }
            if (d.n == 0) {
                if (exec) {
                    d.result = 0;
                } else {
                    perror("Pipe closed unexpectedly");
                    d.result = 1; // Error
                }
                qWarning() << "process finished1" << errno << "str error " << strerror(errno);;
                break;
            }
            qWarning() << "Error reading from pipe";
            d.result = 1; // Error
        }
        close(d.fd[0]);
        qDebug() << "read finished...";
        break;
    }
    }

    delete [] d.argv;
    d.argv = nullptr;
    return true;
}

bool AppLauncher::launchDesktopApp(const QString &desktopFile,
                                   const QStringList &args,
                                   const QString &execPath,
                                   const QStringList &execArgs)
{
    qDebug() << "---execPath---" << execPath;
    GDesktopAppInfo *appInfo = g_desktop_app_info_new_from_filename(desktopFile.toUtf8().constData());
    if (!appInfo) {
        qWarning() << __FUNCTION__ << "appInfo is null";
        return false;
    }

    GList *l = nullptr;
    if (!args.isEmpty()) {
        QString filePath = execArgs.join(" ").contains("sh -c") ? execArgs.join(" ") : args.join(" ");
        QUrl url(filePath);
        QFile file(filePath);

        if (url.isValid() && url.scheme().isEmpty()) {
            if (file.exists() && url.isRelative()) {
                QString fileUriPath = "file://" + filePath;
                char *uri = g_strdup(fileUriPath.toUtf8().constData());
                l = g_list_prepend(l, uri);
            } else {
                qDebug() << filePath << "is not localFile, maybe is command line!";
                return launch(execPath, execArgs);
            }
        } else {
            qDebug() << filePath << " is url type !";
            for (auto uri : args) {
                l = g_list_prepend(l, g_strdup(uri.toUtf8().constData()));
            }
        }
    }
    gboolean ret = g_desktop_app_info_launch_uris_as_manager(appInfo,
                                                             l,
                                                             nullptr,
                                                             G_SPAWN_DEFAULT,
                                                             nullptr,
                                                             nullptr,
                                                             childProcessCallback,
                                                             nullptr,
                                                             nullptr);

    g_list_free(l);
    if (ret == FALSE) {
        return launch(execPath, execArgs);
    }
    return true;
}

QPair<QString, int> AppLauncher::appIdByPid(quint32 pid, ProcessInfo::AppStatus appStatus)
{
    qDebug() << __FUNCTION__ << appStatus << pid;
    auto doFindDesktopFileFunc = [this](int pid)->QPair<QString, int> {
        QString cmdLine = ProcessInfo::cmdline(pid);
        if (cmdLine.isEmpty()) {
            return QPair<QString, int>();
        }
        auto it = m_appsInfo.constBegin();
        while (it != m_appsInfo.constEnd()) {
            if (!it.key().endsWith(".desktop")) {
                ++ it;
                continue;
            }
            auto appList = it.value();
            int index = 0;
            for (const auto &appInfo : appList) {
                QStringList execArgs = appInfo.getArgs();
                QString exec;
                for (auto &subStr : execArgs) { 
                    // 网页版应用Exec中带"xdg-open",cmdline中不带需要去除
                    if (subStr != "xdg-open") {
                        exec.append(subStr);
                        exec.append(" ");
                    }
                }
                if (exec.endsWith(" ")) {
                    exec.remove(exec.size() - 1, 1);
                } 
                // 移除字符串首尾的空格
                exec = exec.trimmed();

                qDebug() << "cmdline" << cmdLine << "exec" << exec;

                if (cmdLine.contains(exec)) {
                    return {it.key(), index};
                }
                // software center
                cmdLine.remove("./");
                if (exec.endsWith(cmdLine)) {
                    return {it.key(), index};
                }
                ++ index;
            }
            ++ it;
        }
        return QPair<QString, int>();
    };

    // 非应用管理拉起的应用
    auto findGlobalDesktopFileFunc = [this](const QString &cmdLine)->QPair<QString, int> {
        auto desktopExecIt = std::find_if(m_desktopExec.constBegin(), m_desktopExec.constEnd(),
                                          [cmdLine](const QPair<QString, QStringList> &exec) {
            QString strExec;
            for (const auto &str : exec.second) {
                strExec.push_back(str);
                strExec.push_back(" ");
            }
            if (strExec.endsWith(" ")) {
                strExec.remove(strExec.size()-1, 1);
            }
            return cmdLine.contains(strExec);
        });
        if (desktopExecIt != m_desktopExec.constEnd()) {
            return { desktopExecIt.key(), -1 };
        }
        return { QString(), -1 };
    };

    QString cmdLine = ProcessInfo::cmdline(pid);
    if (cmdLine.startsWith(kKmreAppMainWindow)) {
        auto appIds = m_appsInfo.keys();
        auto it = std::find_if(appIds.constBegin(), appIds.constEnd(), [&cmdLine](const QString &appId) {
            return cmdLine.contains(appId);
        });
        if (it != appIds.constEnd()) {
            return { *it, 0 };
        }
    }

    // 先匹配 launch 进程
    QPair<QString, int> result;
    ProcessList::iterator dstIt;
    auto it = m_appsInfo.constBegin();
    while (it != m_appsInfo.constEnd()) {
        for (int i=0; i<it->count(); ++i) {
            if (it->at(i).launchPid() == (int)pid) {
                result = { it.key(), i};
                return result;
            }
        }
        ++ it;
    }

    // 根据cmdline进行匹配
    result = doFindDesktopFileFunc(pid);
    if (!result.first.isEmpty()) {
        return result;
    }

    // 匹配父进程、找到最终一个launch（启动）的父进程
    int ppid = ProcessInfo::parentId(pid);
    int appPPid;
    if (ppid > 0) {
        do {
            appPPid = ProcessInfo::parentId(ppid);
            result = doFindDesktopFileFunc(ppid); 
            if (!result.first.isEmpty()) {
                return result;
            } else {
                ppid = appPPid;
            }
        } while (result.first.isEmpty() && appPPid > 0);
    }

    // 非应用管理拉起的应用
    result = findGlobalDesktopFileFunc(cmdLine);

    if (result.first.isEmpty()) {
        cmdLine = ProcessInfo::cmdline(ppid);
        result = findGlobalDesktopFileFunc(cmdLine);
    }

    return result;
}

bool AppLauncher::doActiveProcessByWid(const WindowId &wid)
{
    qDebug() << __FUNCTION__ << wid;
    qint64 pid;
    auto appInfoIt = m_appsInfo.begin();
    while (appInfoIt != m_appsInfo.end()) {
        auto widInfoIt = appInfoIt->begin();
        while (widInfoIt != appInfoIt->end()) {
            if (widInfoIt->wids().contains(wid)) {
                pid = widInfoIt->pid();
                bool ret = updateAppCGroupStatus(pid, ProcessInfo::ForegroundProcess);
                if (ret) {
                    widInfoIt->setCGroup(ProcessInfo::ForegroundProcess);
                    widInfoIt->setLockCgroup(true);
                }
                return true;
            }
            ++ widInfoIt;
        }
        ++ appInfoIt;
    }

    // 如果没找到，或者应用管理重启了
    quint32 windowPid = WindowManager::getPid(wid);
    if (windowPid > 0) {
        m_cgroupInterface->MoveProcess(windowPid, ProcessInfo::ForegroundProcess);
    }

    return true;
}

void AppLauncher::windowChanged(WindowId wid, bool added)
{
    quint32 pid = WindowManager::getPid(wid);
    qDebug() << __FUNCTION__ << wid << pid << added;

    // 接收到该信号时可能进程已经退出了
    if (pid == 0 || !added) {
        auto appIds = m_appsInfo.keys();
        for (const auto &appId : appIds) {
            auto &appList = m_appsInfo[appId];
            for (int i=0; i<appList.count(); ++i) {
                if (appList.at(i).wids().contains(wid)) {
                    appList[i].removeWid(wid);
                    if (appList.at(i).wids().isEmpty()) {
                        appList.removeAt(i);
                        if (appList.isEmpty()) {
                            m_appsInfo.remove(appId);
                        }
                    }
                    qDebug() << __FUNCTION__ << wid << added << m_appsInfo;
                    return;
                }
            }
        }
        return;
    }

    QString cmdLine = ProcessInfo::cmdline(pid);

    auto appStatus = added ? ProcessInfo::Launching : ProcessInfo::Running;
    auto app = appIdByPid(pid, appStatus);
    if (app.first.isEmpty()) {
        if (cmdLine == polkitWindowPidCmdline) {
            m_polkitWindowId = wid;
            qWarning() << "Polkit window pid:" << pid;
        } else {
            qWarning() << "Can not find the desktop file according the pid:" << pid
                       << ProcessInfo::cmdline(pid);
        }
        return;
    }
    qDebug() << "find app id:" << app.first << app.second << pid;
    if (kFreeApps.contains(app.first)) {
        qDebug() << "is free app.";
        return;
    }
    // 通过应用管理打开的应用
    if (app.second > -1) {
        ProcessInfo &appInfo = m_appsInfo[app.first][app.second];
        appInfo.setAppStatus(ProcessInfo::Running);
        if (added) {
            appInfo.setPid(pid);
            appInfo.pushWid(wid);
        } else {
            appInfo.removeWid(wid);
            qDebug() << "wids after remove wid: " << appInfo.wids();
            if (appInfo.wids().isEmpty()) {
                m_appsInfo[app.first].removeAt(app.second);
                if (m_appsInfo.value(app.first).isEmpty()) {
                    m_appsInfo.remove(app.first);
                }
            }
        }
    } else {
        ProcessInfo pi;
        pi.setPid(pid);
        pi.setArgs(QStringList() << ProcessInfo::cmdline(pid));
        pi.setWids(QList<WindowId>() << wid);
        pi.setAppStatus(ProcessInfo::Running);
        if (app.first.endsWith(".desktop")) {
            pi.setDesktopFileName(app.first);
        }

        if (m_appsInfo.contains(app.first)) {
            if (!m_appsInfo.value(app.first).isEmpty()) {
                m_appsInfo[app.first][0] = pi;
            } else {
                m_appsInfo[app.first].push_back(pi);
            }
        } else {
            m_appsInfo[app.first].push_back(pi);
        }
    }

    qDebug() << m_appsInfo;
}

void AppLauncher::moveCachePidToGroup(ProcessInfo::AppCGroup status)
{
    QString pidFilePath = "/sys/fs/cgroup/freezer/CachedProcess/tasks";
    QFile file(pidFilePath);
    if (!file.open(QIODevice::ReadOnly)) {
        qWarning() << "move cached pid faild," << file.errorString();
    }
    auto strPids = file.readAll().split('\n');
    bool ok = false;
    for (const auto &strPid : strPids) {
        if (strPid.isEmpty()) {
            continue;
        }
        int pid = strPid.toInt(&ok);
        if (ok) {
            m_cgroupInterface->MoveProcess(pid, status);
        }
    }
    file.close();
}

void AppLauncher::childProcessCallback(GDesktopAppInfo *appInfo, GPid pid, gpointer userData)
{
    Q_UNUSED(userData)
    QString desktopFile = g_desktop_app_info_get_filename(appInfo);
    if (desktopFile.isEmpty() || !desktopFile.endsWith(".desktop")) {
        return;
    }
    QStringList strList = desktopFile.split("/");
    QString desktopFileName = strList.isEmpty() ? desktopFile : strList.last();
    if (m_appsInfo.contains(desktopFileName)) {
        m_appsInfo[desktopFileName].last().setLaunchPid(pid);
    }
}
